from functools import partialmethod

from django.db import models

from xperms import get_perm_model, enums
from xperms.utils import get_perm


PERM_USER_SLUG = 'users'
PERM_GROUP_SLUG = 'groups'


class PermManagerMetaclass(type):

    def __new__(mcs, name, bases, attrs):
        # adds magical helper methods to the perm manager to filter specified perm type
        new_class = super().__new__(mcs, name, bases, attrs)
        for perm_type in enums.PERM_TYPE_CHOICES:
            setattr(new_class, '{}_perms'.format(perm_type[0]),
                    partialmethod(new_class.TYPE_perms, perm_type=perm_type[0]))

        return new_class


class PermManager(models.Manager, metaclass=PermManagerMetaclass):

    def perm_exists(self, perm, obj=None):
        # determine whether a permission exists
        try:
            get_perm(perm, obj)
        except self.DoesNotExist:
            return False
        return True

    def get_from_str(self, perm, obj=None):
        # get a perm instance based on perm kwargs
        perm_kwargs = get_perm_model().get_perm_kwargs(perm, obj)
        return self.get(**perm_kwargs)

    def _create_from_str(self, perm, obj):
        # create a perm instance based on perm kwargs
        perm_kwargs = get_perm_model().get_perm_kwargs(perm, obj)
        return self.create(**perm_kwargs)

    def create_from_str(self, perms, obj=None):
        if isinstance(perms, str):
            perms = [perms]

        obj_perms = []
        for perm in perms:
            obj_perms.append(self._create_from_str(perm, obj))
        return obj_perms

    def TYPE_perms(self, perm_type):
        # return all perms of the specified type
        # used to generate magical helper methods in the metaclass
        return self.get_queryset().filter(perm_type=perm_type)


class RelatedPermManager(models.Manager):

    def all_perms(self):
        # get all permissions for related group or user
        all_perms_cache_name = 'all_perms_cache'
        if not hasattr(self, all_perms_cache_name):
            perm_pks = set(self.all().values_list('pk', flat=True))
            # If the related object is a user, add permissions from its groups
            if self.query_field_name == PERM_USER_SLUG:
                for group in self.instance.groups.all():
                    perm_pks.update(set(group.perms.all().values_list('pk', flat=True)))

            setattr(self, all_perms_cache_name, self.model.objects.filter(pk__in=perm_pks))
        return getattr(self, all_perms_cache_name)

    def get_perms(self, *perms, obj=None):
        obj_perms = []
        for perm in perms:
            obj_perms.append(get_perm(perm, obj))
        return obj_perms

    def add_perm(self, *perms, obj=None):
        return self.add(*self.get_perms(*perms, obj=obj))

    def remove_perm(self, *perms, obj=None):
        return self.remove(*self.get_perms(*perms, obj=obj))

    def get_perm(self, perm, obj=None):
        # get a permission if it belongs to group or user
        perm = get_perm(perm, obj)

        return self.all_perms().get(pk=perm.pk)

    def has_perm(self, perm, obj=None):
        # determine whether a user or a group has provided permission
        if hasattr(self.instance, 'is_superuser') and self.instance.is_superuser:
            return True

        if perm is None:
            return False
        try:
            self.get_perm(perm, obj)
        except get_perm_model().DoesNotExist:
            return False

        return True



# Generated by Django 3.2.2 on 2022-03-14 09:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('xperms', '0004_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='group',
            name='description',
            field=models.CharField(blank=True, default='', max_length=180),
        ),
    ]



from django.contrib import admin
from django.contrib.admin import ModelAdmin
from django.contrib.admin.views.main import ChangeList

from django.utils.functional import cached_property

from xperms import get_perm_model, enums
from xperms.utils import get_content_type


Perm = get_perm_model()


class Codename:
    ADD = 'add'
    CHANGE = 'change'
    DELETE = 'delete'


class PermAdminMixin:

    @cached_property
    def _perms(self):
        return Perm.objects.filter(content_type=get_content_type(self.model))


class PermChangeList(PermAdminMixin, ChangeList):

    def get_queryset(self, request):
        qs = super().get_queryset(request)

        # just in case there are objects lacking change permission
        if request.user.is_superuser:
            return qs.all()

        object_ids = self._perms.filter(
            object_id__isnull=False,
            users__in=[request.user,],
            codename='change',
        ).values_list('object_id', flat=True)

        return qs.filter(pk__in=object_ids)


class PermModelAdmin(PermAdminMixin, ModelAdmin):

    perms_per_instance = False
    perms_per_instance_author_change = True
    perms_per_instance_author_delete = True

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)

        # add the change and delete permissions for the author of a new model instance
        if not change and self.perms_per_instance:
            for codename in (Codename.CHANGE, Codename.DELETE):
                if hasattr(self, 'perms_per_instance_author_'+codename):
                    self.add_perm(request.user, codename, obj)

    def get_changelist(self, request, **kwargs):
        if self.perms_per_instance:
            return PermChangeList
        else:
            return ChangeList

    def has_change_permission(self, request, obj=None):
        if self.perms_per_instance:
            return self.has_perm(request.user, Codename.CHANGE, obj=obj)

        return self.has_perm(request.user, Codename.CHANGE)

    def has_delete_permission(self, request, obj=None):
        if self.perms_per_instance:
            return self.has_perm(request.user, Codename.DELETE, obj=obj)

        return self.has_perm(request.user, Codename.DELETE)

    def add_perm(self, user, codename, obj):
        object_id = obj.pk if obj is not None else None
        perm_kwargs = {
            'type': enums.PERM_TYPE_OBJECT,
            'content_type': get_content_type(self.model),
            'object_id': object_id,
            'codename': codename,
        }

        perm, _ = Perm.objects.get_or_create(**perm_kwargs)
        user.perms.add(perm)

    def has_perm(self, user, codename=None, obj=None):
        if user.is_superuser:
            return True

        object_id = obj.pk if obj is not None else None

        try:
            perm = self._perms.get(codename=codename, object_id=object_id)
            return user.perms.has_perm(perm)
        except Perm.DoesNotExist:
            return False


class PermAdmin(ModelAdmin):

    list_filter = ['content_type']


admin.site.register(Perm, PermAdmin)



# Generated by Django 2.2.15 on 2020-09-16 18:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('xperms', '0003_migration'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='perm',
            name='groups',
        ),
    ]



# Generated by Django 3.1.1 on 2020-09-16 12:38

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('xperms', '0002_auto_20180713_1450'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AlterField(
            model_name='perm',
            name='type',
            field=models.CharField(choices=[('generic', 'generic'), ('model', 'model'), ('object', 'object')],
                                   default='generic', max_length=10),
        ),
        migrations.AlterField(
            model_name='perm',
            name='users',
            field=models.ManyToManyField(blank=True, related_name='xperms', to=settings.AUTH_USER_MODEL,
                                         verbose_name='users'),
        ),
        migrations.CreateModel(
            name='Group',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('codename', models.CharField(blank=True, max_length=100, null=True, unique=True,
                                              verbose_name='codename')),
                ('name', models.CharField(max_length=255, verbose_name='name')),
                ('xgroups', models.ManyToManyField(blank=True, related_name='parents', to='xperms.Group',
                                                   verbose_name='subgroups')),
                ('xperms', models.ManyToManyField(blank=True, related_name='xgroups', to='xperms.Perm',
                                                  verbose_name='permissions')),
                ('users', models.ManyToManyField(blank=True, related_name='xgroups', to=settings.AUTH_USER_MODEL,
                                                 verbose_name='users')),
            ],
            options={
                'verbose_name': 'permission group',
                'verbose_name_plural': 'permission groups',
                'ordering': ('codename',),
            },
        ),
    ]



# Generated by Django 4.1 on 2022-08-11 13:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('xperms', '0005_group_description'),
    ]

    operations = [
        migrations.AlterField(
            model_name='perm',
            name='type',
            field=models.CharField(choices=[('generic', 'generic'), ('model', 'model'), ('object', 'object'), ('field', 'field')], default='generic', max_length=10),
        ),
    ]



from functools import partialmethod

from django.apps import apps
from django.conf import settings as django_settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.base import ModelBase
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError

from xperms import enums
from xperms.conf import settings
from xperms.exceptions import ObjectNotPersisted, IncorrectContentType, IncorrectObject
from xperms.managers import PermManager, RelatedPermManager
from xperms import get_perm_model


class PermMetaclass(ModelBase):

    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(mcs, name, bases, attrs)
        for perm_type in new_class.PERM_TYPE_CHOICES:
            setattr(new_class, 'is_{}_perm'.format(perm_type[0]),
                    partialmethod(new_class.is_TYPE_perm, perm_type=perm_type[0]))

        return new_class


class BasePerm(models.Model, metaclass=PermMetaclass):

    PERM_TYPE_CHOICES = enums.PERM_TYPE_CHOICES
    PERM_CODENAMES = enums.PERM_CODENAMES

    type = models.CharField(
        max_length=10,
        choices=PERM_TYPE_CHOICES,
        default=enums.PERM_TYPE_GENERIC,
    )
    codename = models.CharField(
        _('codename'),
        max_length=100,
    )
    name = models.CharField(
        _('name'),
        max_length=255,
        null=True,
        blank=True,
    )
    content_type = models.ForeignKey(
        ContentType,
        models.CASCADE,
        verbose_name=_('content type'),
        blank=True,
        null=True,
    )
    object_id = models.TextField(
        _('object pk'),
        null=True,
        blank=True,
    )
    content_object = GenericForeignKey()
    users = models.ManyToManyField(
        django_settings.AUTH_USER_MODEL,
        verbose_name=_('users'),
        related_name='xperms',
        blank=True
    )

    objects = PermManager()

    class Meta:
        abstract = True
        verbose_name = _('permission')
        verbose_name_plural = _('permissions')
        ordering = ('codename',)
        unique_together = (
            ('type', 'codename', 'content_type', 'object_id', 'field_name'),
        )

    def __str__(self):
        if self.name:
            return self.name

        permission_name = str(self.PERM_CODENAMES.get(self.codename, self.codename))
        name = getattr(self, '_{}_perm_name'.format(self.type), '')

        return ' | '.join(('Permission', name, permission_name))

    @cached_property
    def model(self):
        return self.content_type.model_class()

    @classmethod
    def get_perm_kwargs(cls, perm, obj=None):
        perm_type, perm_arg_string = perm.split('.', 1)

        model = content_type = object_id = None

        if perm_type == enums.PERM_TYPE_MODEL or perm_type == enums.PERM_TYPE_OBJECT:
            model_name, codename = perm_arg_string.rsplit('.', 1)
            model = apps.get_model(model_name)
            if perm_type == enums.PERM_TYPE_OBJECT:
                if not isinstance(obj, models.Model):
                    raise IncorrectObject(_('Object {} must be a model instance').format(obj))
                if not isinstance(obj, model):
                    raise IncorrectContentType(_('Object {} does not have a correct content type').format(obj))
                if obj.pk is None:
                    raise ObjectNotPersisted(_('Object {} needs to be persisted first').format(obj))

                object_id = obj.pk
        else:
            codename = perm_arg_string

        if model:
            content_type = ContentType.objects.get_for_model(model)

        return dict(
            type=perm_type,
            codename=codename,
            content_type=content_type,
            object_id=object_id
        )

    def get_wildcard_perm(self):
        return self.objects.filter(
            type=self.type,
            codename=enums.PERM_CODENAME_WILDCARD,
            content_type=self.content_type,
            object_id=self.object_id
        )

    def is_TYPE_perm(self, perm_type):
        return self.type == perm_type

    @property
    def _model_perm_name(self):
        return _('model {}').format(
            self.model.__name__
        )

    @property
    def _object_perm_name(self):
        return _('model {} | object {}').format(
            self.model.__name__,
            self.content_object,
        )


class Group(models.Model):

    codename = models.CharField(
        _('codename'),
        max_length=100,
        unique=True,
        null=True,
        blank=True
    )
    name = models.CharField(
        _('name'),
        max_length=255,
        null=False,
        blank=False,
    )
    xperms = models.ManyToManyField(
        settings.PERM_MODEL,
        verbose_name=_('permissions'),
        related_name='xgroups',
        blank=True
    )
    xgroups = models.ManyToManyField(
        'Group',
        verbose_name=_('subgroups'),
        blank=True,
        related_name='parents',
    )
    users = models.ManyToManyField(
        django_settings.AUTH_USER_MODEL,
        verbose_name=_('users'),
        related_name='xgroups',
        blank=True
    )
    description = models.CharField(max_length=180, default='', blank=True)

    class Meta:
        verbose_name = _('permission group')
        verbose_name_plural = _('permission groups')
        ordering = ('codename',)

    def __str__(self):
        return self.name



from xperms.conf import settings
from django.utils.translation import gettext_lazy as _


PERM_CODENAME_ADD = 'add'
PERM_CODENAME_CHANGE = 'change'
PERM_CODENAME_DELETE = 'delete'
PERM_CODENAME_WILDCARD = '*'

DEFAULT_PERM_CODENAMES = {
    PERM_CODENAME_ADD: _('add'),
    PERM_CODENAME_CHANGE: _('change'),
    PERM_CODENAME_DELETE: _('delete'),
    PERM_CODENAME_WILDCARD: _('wildcard'),
}

PERM_CODENAMES = dict(DEFAULT_PERM_CODENAMES, **settings.PERM_CODENAMES)

PERM_TYPE_GENERIC = 'generic'
PERM_TYPE_MODEL = 'model'
PERM_TYPE_OBJECT = 'object'
PERM_TYPE_FIELD = 'field'

DEFAULT_PERM_TYPE_CHOICES = (
    (PERM_TYPE_GENERIC, _('generic')),
    (PERM_TYPE_MODEL, _('model')),
    (PERM_TYPE_OBJECT, _('object')),
    (PERM_TYPE_FIELD, _('field')),
)

PERM_TYPE_CHOICES = DEFAULT_PERM_TYPE_CHOICES + settings.PERM_TYPE_CHOICES



# Generated by Django 2.0 on 2018-07-13 14:50

from django.db import migrations, models
import django.db.models.manager


class Migration(migrations.Migration):

    dependencies = [
        ('xperms', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='perm',
            options={'base_manager_name': 'related_manager'},
        ),
        migrations.AlterModelManagers(
            name='perm',
            managers=[
                ('related_manager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterField(
            model_name='perm',
            name='object_id',
            field=models.TextField(blank=True, null=True, verbose_name='object pk'),
        ),
        migrations.AlterUniqueTogether(
            name='perm',
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name='perm',
            name='field_name',
        ),
    ]



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase, Client

from articles.models import Article

from .factories import UserFactory, GroupFactory


class ArticleTestCase(TestCase):

    def setUp(self):
        self.user = UserFactory()
        self.client = Client()

    def _get_content_type(self):
        return ContentType.objects.get_for_model(Article)


class ArticleUserPermTestCase(ArticleTestCase):
    pass


class ArticleGroupPermTestCase(ArticleTestCase):

    def setUp(self):
        super().setUp()
        self.group = GroupFactory()
